import json
import logging
import os
import uuid
from typing import Union, List, Tuple
import time

from .modifier import Modifier
from .logger import cleanup_logger, universal_logger_init, twb_logger
from .parallelization import RDSProcessManager, RDSProcessController
from .utils import get_file_list, decompress_zstd, prepare_output_dir, get_curr_version, get_line_positions, \
    cleanup_dir, read_line_in_file, compress_zstd, COMPRESSION_EXTENSION, parse_schema

_DEFAULT_NUM_PROC = 1
_DEFAULT_LOG_LEVEL = logging.DEBUG


class Reader:
    """
    The core class to read the generated temporal wikipedia blocks.

    Basic usage:

        import twb
        reader = twb.Reader()  # Create an instance.
        reader.preload('./test/sample_data/minimal_sample.xml')  # Preload the files to be processed.

    You must preload the files to be processed before building the blocks.
    Preload function accepts a path to a file or a directory, and can be called multiple times (for multiple files).

    Attributes:
        num_proc (int): The number of processes to use.
        log_dir (str): The dir to the log file.
        log_level (int): The log level.
        files (list): A list of files to be read.
        modifiers (list): A list of modifiers to be applied.
    """

    def __init__(self,
                 log_dir: Union[str, None] = None,
                 num_proc: int = _DEFAULT_NUM_PROC,
                 log_level: int = _DEFAULT_LOG_LEVEL):
        """
        :param log_dir: the dir to the log file (default: None)
        :param num_proc: the number of processes to use (default: 1)
        :param log_level: the log level (default: logging.INFO)
        """
        self.log_dir = log_dir
        self.num_proc = num_proc
        self.log_level = log_level

        self.files = []
        self.modifiers = []

        # If log dir exists, remove it first.
        cleanup_logger(log_name='reader', log_dir=log_dir)

        # Initialize the logger.
        universal_logger_init(log_name='reader', log_dir=log_dir, log_level=log_level)

    def preload(self, path: str):
        """
        Preload the files to be processed.
        It will not actually load to the memory until the build() method is called.
        :param path: the path of a file or a directory
        :raise ValueError: if the path is empty
        :raise FileNotFoundError: if the path does not exist
        """
        if not path:
            raise ValueError('The path cannot be empty.')
        if not os.path.exists(path):
            raise FileNotFoundError('The path does not exist.')
        self.files.extend(get_file_list(path))

    def glimpse(self) -> Union[Tuple[dict, dict], Tuple[None, None]]:
        """
        Take a glimpse of the data.
        It could still be large if one object contains a lot of information (e.g. many revisions, long article).
        """
        if len(self.files) == 0:
            twb_logger.warning('No file is loaded.')
            return None

        # Randomly select a file.
        picked_index = int(uuid.uuid4().int % len(self.files))
        glimpse_path = self.files[picked_index]

        twb_logger.info(f'Randomly chosen file: {glimpse_path}')

        # Prepare the temporary directory for glimpse.
        glimpse_temp_dir = os.path.join(os.getcwd(), '.glimpse')
        os.makedirs(glimpse_temp_dir, exist_ok=True)

        twb_logger.info(f'Decompressing...')

        # Decompress the file.
        decompressed_path = _decompress_executor(glimpse_path, glimpse_temp_dir)
        first_block_text = read_line_in_file(decompressed_path, 0).rstrip('\n')

        if first_block_text[0] != '{' or first_block_text[-1] != '}':
            twb_logger.error(f'Invalid starting of block or end of block.')
            return None, None

        # Read the first block into memory and then delete the decompressed file.
        first_block = json.loads(first_block_text)
        cleanup_dir(glimpse_temp_dir)

        twb_logger.info(f'Glimpse finished.')

        return first_block, parse_schema(first_block)

    def decompress(self, output_dir: str):
        """
        Decompress the selected files.
        :param output_dir: the directory to store the decompressed files
        """
        # Log the version.
        twb_logger.info(f'TWB Package Version: {get_curr_version()}')

        # Prepare the output directory.
        prepare_output_dir(output_dir)

        start_time = time.time()

        pm = RDSProcessManager(
            log_name='decompress',
            log_dir=self.log_dir,
            log_level=self.log_level,
            num_proc=self.num_proc
        )
        for file in self.files:
            pm.apply_async(
                executable=_decompress_executor,
                args=(file, output_dir),
                use_controller=False
            )

        pm.close()
        pm.join()

        end_time = time.time()
        execution_duration = end_time - start_time
        twb_logger.info(f'Decompression finished. (Duration: {execution_duration:.2f}s)')

    def add_modifier(self, modifier: Modifier):
        """
        Map a function to each block.
        :param modifier: the modifier to be added
        """
        self.modifiers.append(modifier)

    def build_modification(self, output_dir: str):
        """
        Build the blocks after apply the modifiers.
        :param output_dir: the directory to store the modified files
        """
        # Log the version.
        twb_logger.info(f'TWB Package Version: {get_curr_version()}')

        start_time = time.time()

        # Prepare the output directory.
        prepare_output_dir(output_dir)

        curr_count = 0
        total_count = len(self.files)

        global_temp_dir = os.path.join(output_dir, 'temp')
        os.makedirs(global_temp_dir, exist_ok=True)

        def _success_callback(*args):
            pass

        def _error_callback(e):
            twb_logger.error(f'Error occurred when processing block: {e}')

        for file_path in self.files:
            twb_logger.info(f'Start: {file_path}')

            file_start_time = time.time()

            # Create temporary directory for this file.
            temp_id = uuid.uuid4().hex
            temp_dir = os.path.join(output_dir, 'temp', temp_id)
            os.makedirs(temp_dir, exist_ok=True)

            decompression_temp_dir = os.path.join(temp_dir, 'decompression')
            os.makedirs(decompression_temp_dir, exist_ok=True)
            compression_temp_dir = os.path.join(temp_dir, 'compression')
            os.makedirs(compression_temp_dir, exist_ok=True)

            # Decompress the file first (this step will not be done in parallel because this is faster).
            twb_logger.debug(f'Decompressing...')
            decompressed_path = _decompress_executor(file_path, decompression_temp_dir)

            # Compute target path. The target path should be within compression temporary directory.
            target_filename = os.path.basename(decompressed_path)
            target_path = os.path.join(compression_temp_dir, target_filename)

            line_positions = get_line_positions(decompressed_path)

            twb_logger.info(f'Start modifying {len(line_positions)} blocks...')

            # Get line positions for all files.
            pm = RDSProcessManager(
                log_name='reader',
                log_dir=self.log_dir,
                log_level=self.log_level,
                num_proc=self.num_proc
            )

            for line_position in line_positions:
                pm.apply_async(
                    executable=_modify_executor,
                    args=(decompressed_path, line_position, target_path, self.modifiers),
                    use_controller=True,
                    callback=_success_callback,
                    error_callback=_error_callback
                )

            pm.close()
            pm.join()

            twb_logger.info(f'Finished modifying {len(line_positions)} blocks. Compressing...')

            # Compress the file.
            output_path = os.path.join(output_dir, os.path.basename(target_path) + COMPRESSION_EXTENSION)
            compress_zstd(target_path, output_path)

            twb_logger.info(f'Finished compressing. Cleaning up...')

            # Clean up the temporary directory at this step.
            cleanup_dir(temp_dir)

            twb_logger.info(f'Finished cleaning up.')

            file_end_time = time.time()
            file_execution_duration = file_end_time - file_start_time

            twb_logger.info(f'Finished: {file_path} -- (took {file_execution_duration:.2f}s)')

            # Log the progress.
            curr_count += 1
            twb_logger.info(f'Progress: {curr_count} / {total_count} = ({curr_count / total_count * 100:.2f}%)')

        # Clean up the global temporary directory.
        cleanup_dir(global_temp_dir)

        end_time = time.time()
        execution_duration = end_time - start_time

        # Log the end of the task.
        twb_logger.info(f'All done! Finished all files. (took {execution_duration:.2f}s in total)')


def _decompress_executor(path: str, output_dir: str) -> str:
    original_name = os.path.split(path)[1]
    decompressed_name = original_name.replace('.zst', '')
    decompressed_path = os.path.join(output_dir, decompressed_name)
    decompress_zstd(path, decompressed_path)
    return decompressed_path


def _line_position_processor(path: str, output_dir: str):
    """
    The processor for the line position process.
    """
    temp_id = uuid.uuid4().hex

    temp_dir = os.path.join(output_dir, 'temp', temp_id)
    os.makedirs(temp_dir, exist_ok=True)

    _decompress_executor(path, temp_dir)

    return path, get_line_positions(path)


def _modify_executor(controller: RDSProcessController,
                     path: str,
                     position: int,
                     target_path: str,
                     modifiers: List[Modifier]):
    """
    The executor for the modification process.
    :param controller: the controller of the process
    :param path: the path of the file to be processed
    :param position: the position of the file to be processed
    :param target_path: the directory to store the modified files
    :param modifiers: the list of modifiers to be applied
    """
    start_time = time.time()

    controller.logdebug(f'Processing block: {position}')
    block_text = read_line_in_file(path, position).rstrip('\n')
    if block_text[0] != '{' or block_text[-1] != '}':
        controller.logerr(f'Invalid starting of block or end of block: {path} @ {position}')
        return

    # Parse the block from text to JSON.
    block = json.loads(block_text)

    # Apply the modifiers.
    for modifier in modifiers:
        if block is None:
            break
        block = modifier.modify(block)

    # We write the output only if the block is not None. Otherwise, we remove this block.
    if block is not None:
        with controller.parallel_lock:
            logging.debug(f'Storing JSONL to {target_path}...')
            try:
                with open(target_path, 'a', buffering=1) as f:
                    f.write(json.dumps(block) + '\n')
            except Exception as e:
                controller.logerr(f'Error occurred when writing block to file: {e}')
            logging.debug(f'Storing JSONL done: {target_path}.')
    else:
        controller.logdebug(f'Removed block because of "None": {position}')

    end_time = time.time()
    execution_duration = end_time - start_time

    controller.logdebug(f'Finished block: {position} -- (took {execution_duration:.2f}s)')



import logging
import os
import time
from typing import Union

import py7zr

from .logger import universal_logger_init, twb_logger, cleanup_logger
from .parallelization import RDSProcessManager, RDSProcessController
from .utils import get_file_list, decompress_zstd, get_curr_version, prepare_output_dir

_DEFAULT_NUM_PROC = 1
_DEFAULT_LOG_LEVEL = logging.INFO


class Decompressor:
    """
    The core class to decompress normal 7z files or zstd files in cluster environment.

    NOTE: Decompressor is used for decompressing a file or a directory of files in cluster environment where
    command line decompression is not available. This class is not used to read the dataset.
    If you are trying to read the dataset, please use `twb.Reader` rather than `twb.Decompressor`.

    You must preload the files to be processed before starting decompressing.
    Preload function accepts a path to a file or a directory, and can be called multiple times (for multiple files).

    Attributes:
        num_proc (int): The number of processes to use.
        log_dir (str): The dir to the log file.
        log_level (int): The log level.
        files (list): A list of files to be read.
    """

    def __init__(self,
                 log_dir: Union[str, None] = None,
                 num_proc: int = _DEFAULT_NUM_PROC,
                 log_level: int = _DEFAULT_LOG_LEVEL):
        """
        :param log_dir: the dir to the log file (default: None)
        :param num_proc: the number of processes to use (default: 1)
        :param log_level: the log level (default: logging.INFO)
        """
        self.log_dir = log_dir
        self.num_proc = num_proc
        self.log_level = log_level

        self.files = []

        # If log dir exists, remove it first.
        cleanup_logger(log_name='decompressor', log_dir=log_dir)

        # Initialize the logger.
        universal_logger_init(log_name='decompressor', log_dir=log_dir, log_level=log_level)

    def preload(self, path: str):
        """
        Preload the files to be processed.
        It will not actually load to the memory until the go() method is called.
        :param path: the path of a file or a directory
        :raise ValueError: if the path is empty
        :raise FileNotFoundError: if the path does not exist
        """
        if not path:
            raise ValueError('The path cannot be empty.')
        if not os.path.exists(path):
            raise FileNotFoundError('The path does not exist.')
        self.files.extend(get_file_list(path))

    def start(self, output_dir: str):
        """
        Decompress the files.
        :param output_dir: the output directory
        """
        # Log the version.
        twb_logger.info(f'TWB Package Version: {get_curr_version()}')

        num_proc = self.num_proc
        log_dir = self.log_dir
        log_level = self.log_level

        # Prepare the output directory.
        prepare_output_dir(output_dir)

        start_time = time.time()

        pm = RDSProcessManager(
            log_name='decompressor',
            log_dir=log_dir,
            log_level=log_level,
            num_proc=num_proc
        )

        # TODO: Add a error handler to handle the error when decompressing a file or process terminated unexpectedly.

        for file in self.files:
            pm.apply_async(
                executable=_decompress_file,
                args=(file, output_dir),
                use_controller=True
            )

        pm.close()
        pm.join()

        end_time = time.time()
        execution_duration = end_time - start_time
        twb_logger.info(f'Decompression finished in {execution_duration:.2f} seconds.')


def _decompress_file(controller: RDSProcessController, path: str, output_dir: str):
    start_time = time.time()
    original_name = os.path.split(path)[1]
    if original_name.endswith('.zst'):
        decompressed_name = original_name.replace('.zst', '')
        decompressed_path = os.path.join(output_dir, decompressed_name)
        decompress_zstd(path, decompressed_path)
    elif original_name.endswith('.7z'):
        decompressed_name = original_name.replace('.7z', '')
        decompressed_path = os.path.join(output_dir, decompressed_name)
        with py7zr.SevenZipFile(path, mode='r', mp=False) as z:
            z.extractall(path=decompressed_path)
    else:
        controller.logwarn(f'Unknown file type: {original_name}')
        return
    end_time = time.time()
    execution_duration = end_time - start_time
    controller.logdebug(f'Decompression took {execution_duration:.2f} seconds. ({original_name})')



import logging
import os
import time
from typing import List, Union, Callable
import xmltodict
import py7zr
import json
import uuid

from .logger import universal_logger_init, twb_logger, cleanup_logger
from .utils import get_file_list, compress_zstd, get_memory_consumption, cleanup_dir, get_curr_version, \
    prepare_output_dir, COMPRESSION_EXTENSION
from .bip import BlockInteriorProcessor, DefaultBIP
from .parallelization import RDSProcessManager, RDSProcessController

_DEFAULT_REVISIONS_PER_BLOCK = 1000000
_DEFAULT_START_INDEX = 0
_DEFAULT_NUM_PROC = 1
_DEFAULT_LOG_LEVEL = logging.DEBUG


class Builder:
    """
    The core class to generate the blocks from Wikipedia Edit History chunk.

    Basic usage:

        import twb
        builder = twb.Builder()  # Create an instance.
        builder.preload('./test/sample_data/minimal_sample.xml')  # Preload the files to be processed.
        builder.build('./test/output')  # Build the blocks.

    You must preload the files to be processed before building the blocks.
    Preload function accepts a path to a file or a directory, and can be called multiple times (for multiple files).

    Attributes:
        num_proc (int): The number of processes to use.
        log_dir (str): The dir to the log file.
        log_level (int): The log level.
        revisions_per_block (int): The number of revisions per block.
        files (List[str]): The list of files to be processed. It will be finalized when the build() method is called.
    """

    def __init__(self,
                 log_dir: Union[str, None] = None,
                 num_proc: Union[int, None] = _DEFAULT_NUM_PROC,
                 revisions_per_block: int = _DEFAULT_REVISIONS_PER_BLOCK,
                 log_level: int = _DEFAULT_LOG_LEVEL):
        """
        :param log_dir: the dir to the log file (default: None) (Will be cleaned up if exists)
        :param num_proc: the number of processes (default: 1) (Set to None to use all available processes)
        :param revisions_per_block: the number of revisions per block (default: 300000)
        :param log_level: the log level (default: logging.DEBUG)
        """
        self.log_dir = log_dir
        self.num_proc = num_proc
        self.revisions_per_block = revisions_per_block
        self.log_level = log_level

        self.files = []

        # Clean up the log dir if exists.
        cleanup_logger(log_name='builder', log_dir=log_dir)

        # Initialize the logger.
        universal_logger_init(log_name='builder', log_dir=log_dir, log_level=log_level)

    def preload(self, path: str):
        """
        Preload the files to be processed.
        It will not actually load to the memory until the build() method is called.
        :param path: the path of a file or a directory
        :raise ValueError: if the path is empty
        :raise FileNotFoundError: if the path does not exist
        """
        if not path:
            raise ValueError('The path cannot be empty.')
        if not os.path.exists(path):
            raise FileNotFoundError('The path does not exist.')
        self.files.extend(get_file_list(path))

    def build(self,
              output_dir: str,
              start_index: int = _DEFAULT_START_INDEX,
              processor: BlockInteriorProcessor = DefaultBIP(),
              compress: bool = True):
        """
        Build the blocks.
        :param output_dir: the output directory for the blocks (will be created if not exists)
        :param start_index: the starting index of the blocks (default: 0)
        :param processor: the interior processor for blocks (default: DefaultBIP)
        :param compress: whether to compress the blocks (default: True)
        :raise Warning: if there is no file to process
        """
        # Log the version.
        twb_logger.info(f'TWB Package Version: {get_curr_version()}')

        log_dir = self.log_dir
        num_proc = self.num_proc
        revisions_per_block = self.revisions_per_block
        log_level = self.log_level
        zip_file_list = self.files

        # If there is no file to process, raise a warning and return.
        if len(zip_file_list) == 0:
            raise Warning('There is no file to process.')

        # Prepare the output directory.
        prepare_output_dir(output_dir)

        # Create the temporary directory.
        temp_dir = os.path.join(output_dir, 'temp')
        os.makedirs(temp_dir)

        # Decompress the files in parallel.
        twb_logger.info('Starting RDS process manager...')

        process_manager_context = {
            'processor': processor,  # The interior processor for blocks.
            'compress': compress,  # Whether to compress the blocks.
            'revisions_per_block': revisions_per_block,  # Number of revisions per block.
            'total_count': len(zip_file_list),  # Total number of files to process.
        }

        curr_count = 0
        total_count = len(zip_file_list)

        def _success_callback(result):
            nonlocal curr_count
            curr_count += 1
            curr_processed_progress = curr_count / total_count * 100 if total_count > 0 else -1
            twb_logger.info(f'{curr_count} / {total_count} ==> ({curr_processed_progress:.2f}%)')

        def _error_callback(e):
            twb_logger.error('An error occurred in a sub-process which makes it terminated.',
                             'Check next error log for details.')
            twb_logger.error(e)

            nonlocal curr_count
            curr_count += 1
            curr_processed_progress = curr_count / total_count * 100 if total_count > 0 else -1
            twb_logger.info(f'{curr_count} / {total_count} ==> ({curr_processed_progress:.2f}%) (Terminated)')

        start_time = time.time()

        pm = RDSProcessManager(
            log_name='builder',
            log_dir=log_dir,
            log_level=log_level,
            num_proc=num_proc,
            start_index=start_index
        )
        for path in zip_file_list:
            pm.apply_async(
                executable=_file_processor,
                args=(path, output_dir, process_manager_context),
                callback=_success_callback,
                error_callback=_error_callback,
            )

        pm.close()
        pm.join()

        end_time = time.time()
        execution_duration = end_time - start_time
        twb_logger.info(f"!!! All done. (Took {execution_duration:.2f} seconds in total)")

        # Clean up the temporary directory.
        cleanup_dir(temp_dir)


def _file_processor(controller: RDSProcessController,
                    path: str,
                    output_dir: str,
                    context: dict):
    # If the file does not exist, skip it.
    if not os.path.exists(path):
        controller.logfatal(f'File does not exist: {path}')
        return

    # If the file is not a 7z file, skip it.
    if not path.endswith('.7z'):
        controller.logwarn(f'Skipped because the file is not a 7z file. ({path})')
        return

    # Get the block interior processor from the running context.
    block_interior_processor = context.get('processor', DefaultBIP())
    should_compress = context.get('compress', True)
    revisions_per_block = context.get('revisions_per_block', _DEFAULT_REVISIONS_PER_BLOCK)

    # Generate a temporary id for the file.
    temp_id = uuid.uuid4().hex

    # Compute the path for the temporary directory in this task (associated with this temp_id).
    temp_dir = os.path.join(output_dir, 'temp', temp_id)

    # Register the temporary directory for cleanup previous temporary folders associated with the same pid.
    controller.register(temp_dir)

    # Create the temporary directory.
    os.makedirs(temp_dir)

    # Create the temporary directories for compression and decompression.
    compression_temp_dir = os.path.join(temp_dir, 'compression_temp')
    os.makedirs(compression_temp_dir)
    decompression_temp_dir = os.path.join(temp_dir, 'decompression_temp')
    os.makedirs(decompression_temp_dir)

    # If the temporary directories are not created properly, skip it.
    if not os.path.exists(compression_temp_dir) or not os.path.exists(decompression_temp_dir):
        controller.logfatal(f'Failed to create temporary directories for archive: {path}')
        return

    # Log the start time.
    very_start_time = time.time()

    # First try block, catching issues during decompression.
    try:
        # Decompress the file. It has multi-threading support built-in and enabled by default.
        with py7zr.SevenZipFile(path, mode='r', mp=False) as z:
            start_time = time.time()
            z.extractall(path=decompression_temp_dir)
            end_time = time.time()
            execution_duration = end_time - start_time
            archive_filename = os.path.basename(path)
            controller.logdebug(f'Decompression took {execution_duration:.2f} seconds. ({archive_filename})')
        decompressed_files = get_file_list(decompression_temp_dir)
    except Exception as e:
        controller.logerr(f'Decompression failed at {path} with error:', e)
        cleanup_dir(temp_dir)
        return

    # Second try block, catching issues during processing (XML parsing, etc).
    try:
        def get_new_output_path():
            """
            Get the path of the next output file.
            """
            nonlocal controller, compression_target_dir
            controller.logdebug('Ready to declare index (cp).')
            curr_index = str(controller.declare_index()).zfill(8)  # 8 digits in total for scalability.
            controller.logdebug('Done declaring index (cp).')
            return os.path.join(compression_target_dir, f'block_{curr_index}.jsonl')

        compression_target_dir = compression_temp_dir if should_compress else output_dir

        total_article_count = 0
        total_block_count = 1
        total_revision_count = 0

        revision_count = 0
        memory_usage_records = []
        all_memory_usage_records = []
        curr_output_path = get_new_output_path()

        def _super_callback(article: dict):
            """
            The super callback for the XML parser. It will be called recursively for each article.
            """
            controller.logdebug('_super_callback() calling.')

            nonlocal total_article_count, total_block_count, total_revision_count
            nonlocal revision_count, memory_usage_records, all_memory_usage_records, curr_output_path

            controller.logdebug('_super_callback() started.')

            # If previous batch is full, store the max memory usage and reset the counters.
            if revision_count >= revisions_per_block:
                # Store the max memory usage in previous batch.
                all_memory_usage_records.append(max(memory_usage_records))
                # Update the block count.
                total_block_count += 1
                # Reset the counters.
                revision_count = 0
                memory_usage_records = []
                curr_output_path = get_new_output_path()

            # Compute the number of revisions in the current article.
            if 'revision' in article and type(article['revision']) is list:
                curr_article_revision_count = len(article['revision'])
            else:
                curr_article_revision_count = 1

            # Update the counters.
            revision_count += curr_article_revision_count
            total_revision_count += curr_article_revision_count
            total_article_count += 1

            controller.logdebug(f'Total revision count: {total_revision_count}')

            # Store the article into the current output file.
            _store_article_to_jsonl(article=article, output_path=curr_output_path)

            controller.logdebug(f'Stored.')

            # Record the memory usage.
            memory_usage_records.append(get_memory_consumption())

            controller.logdebug(f'Memory: {memory_usage_records[-1]} MB')

        controller.logdebug(f'Big parsing loop started. (revisions_per_block={revisions_per_block})')

        # We store articles into JSONL files in a streaming manner, along the way of parsing XML files.
        # Therefore, we don't have to keep all the results in the memory, which is a huge problem.
        for path in decompressed_files:
            _parse_xml(xml_path=path, processor=block_interior_processor, super_callback=_super_callback)

        controller.logdebug('Big parsing loop done.')

        if len(memory_usage_records) > 0:
            all_memory_usage_records.append(max(memory_usage_records))

        controller.logdebug(f'Parsing done. {total_article_count} articles processed.',
                            f'({total_revision_count} revisions are processed into {total_block_count} blocks)',
                            f'(Highest Memory: {max(all_memory_usage_records)} MB)')

        cleanup_dir(decompression_temp_dir)  # Delete temporary files used for decompression.

        controller.logdebug(f'Cleaning up decompression temp dir done.')

        if should_compress:
            # Store the results to JSONL files.
            json_files = get_file_list(compression_temp_dir)

            controller.logdebug(f'Compression started. {len(json_files)} files to compress.')

            # Compress the files.
            for json_file in json_files:
                controller.logdebug(f'Compressing {json_file}...')
                _compress_file(path=json_file, output_dir=output_dir, controller=controller)

            controller.logdebug(f'Compression done. {len(json_files)} files compressed in total.')

    except Exception as e:
        controller.logerr(f'Parsing failed at {path}:', e)
    finally:
        # Clean up the temporary directory.
        cleanup_dir(temp_dir)
        controller.logdebug(f'Temporary directory cleaned up: {temp_id}.')

        very_end_time = time.time()
        total_execution_duration = very_end_time - very_start_time
        controller.loginfo(f'File done: {archive_filename}. (Duration: {total_execution_duration:.2f}s)')

        controller.unregister()


def _parse_xml(xml_path: str, processor: BlockInteriorProcessor, super_callback: Callable[[dict], None]):
    """
    Parse the XML file into a list of JSON objects.
    :param xml_path: the path of the XML file
    :param processor: the interior processor for blocks
    :return: the list of parsed results
    """
    def _inner_callback(path, item):
        nonlocal super_callback

        tag_name = path[-1][0]
        item_type = type(item)

        # If the item is a page, we don't need to process it.
        if tag_name != 'page':
            return True

        # If the item is not a dictionary, it means that the item is a leaf node, and we don't expect it to be a block.
        if item_type is not dict:
            return True

        logging.debug(f'Memory: {get_memory_consumption()} MB')

        logging.debug(f'Start callback: <{tag_name}>.')
        super_callback(item)
        logging.debug(f'Callback done: <{tag_name}>.')

        return True

    with open(xml_path, 'rb') as xml_file:
        xmltodict.parse(
            xml_file,
            item_depth=processor.read_depth,
            item_callback=_inner_callback,
        )

        logging.debug(f'Parsing done: {xml_path}')


def _store_article_to_jsonl(article: dict, output_path: str):
    """
    Store the blocks to a JSONL file.
    :param article: the article to store
    :param output_path: the path to store the JSONL file
    :return: the list of JSONL file paths
    """
    logging.debug(f'Dumping JSONL for {output_path}...')
    dumped_article_jsonl = json.dumps(article) + '\n'
    logging.debug(f'Storing JSONL (len: {len(dumped_article_jsonl)}) to {output_path}...')
    try:
        with open(output_path, 'a', buffering=10 * 1024 * 1024) as f:
            logging.debug(f'Writing to {output_path}...')
            f.write(dumped_article_jsonl)
    except Exception as e:
        logging.fatal(f'Error in writing article: {e}')
    logging.debug(f'Storing JSONL done: {output_path}.')


def _compress_file(path: str, output_dir: str, controller: RDSProcessController):
    """
    Compress a file.
    :param path: the path of the file to compress
    :param output_dir: the directory to store the compressed file
    :return:
    """
    # If the file is not a JSONL file, skip it.
    if not path.endswith('.jsonl'):
        controller.logwarn(f'Skipped because the file is not a JSONL file. ({path})')
        return

    output_path = os.path.join(output_dir, os.path.basename(path) + COMPRESSION_EXTENSION)

    try:
        # Compress the file using Zstandard.
        # TODO: Might be able to specify multiprocessing or multithreading. Check the documentation.
        #  If possible, we should use multithreading in order to utilize the CPU cores in this case.
        compress_zstd(input_path=path, output_path=output_path)
    except Exception as e:
        controller.logerr(f'Compression failed at {path}:', e)



import logging
from abc import ABC, abstractmethod
from typing import Union

_DEFAULT_NUM_PROC = 1
_DEFAULT_LOG_LEVEL = logging.INFO


class Modifier(ABC):
    """
    The core class to define how to modify the JSON content.
    """
    @abstractmethod
    def modify(self, content: dict) -> Union[dict, None]:
        """
        Returns a list of batches of URLs to download.
        :param content: The JSON content to be modified.
        :return: Modified JSON content. Return None if the content should be removed.
        """
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['twb']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.12.2,<5.0.0',
 'jsonlines>=3.1.0,<4.0.0',
 'psutil>=5.9.5,<6.0.0',
 'py7zr>=0.20.5,<0.21.0',
 'requests>=2.28.2,<3.0.0',
 'xmltodict>=0.13.0,<0.14.0',
 'zstandard>=0.21.0,<0.22.0']

setup_kwargs = {
    'name': 'twb-project',
    'version': '0.7.3a15',
    'description': 'An unified tool for the research in extracting information from Wikipedia Edit History chunk.',
    'long_description': '<img src="https://imagedelivery.net/Dr98IMl5gQ9tPkFM5JRcng/49178640-2f6d-4c23-e56f-a48eca531200/HD" alt="TWB" />\n\n# Temporal Wikipedia Blocks (TWB)\n\nTemporal Wikipedia Blocks (TWB) is a powerful Python package designed to process the extensive edit history of Wikipedia pages into easily manageable and memory-friendly blocks. The package is specifically developed to enable efficient parallelization and composition of these blocks to facilitate faster processing and analysis of large Wikipedia datasets. The original design of this package is to build other Wikipedia-oriented datasets on top of it.\n\nThe package works by dividing the Wikipedia edit history into temporal blocks, which are essentially subsets of the complete dataset that are based on time intervals. These blocks can then be easily processed and analyzed without the need to load the entire dataset into memory.\n\n## Installation\n\nThe package is available on PyPI and can be installed using pip:\n\n```bash\npip install twb-project\n```\n\n## Benefits\n\n- **Efficient**: The package is designed to be memory-friendly and can be easily parallelized to process large datasets.\n- **Fast**: The package is designed to be fast and can be easily optimized to process large datasets.\n- **Flexible**: The package is designed to be flexible and can be easily extended to support other types of blocks.\n- **Composable**: The package is designed to be composable and can be easily combined with other packages to build other datasets.\n\n## Specification\n\n- Default compression method: ZStandard.\n',
    'author': 'Lingxi Li',
    'author_email': 'hi@lingxi.li',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://twb.lingxi.li/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



import os
from typing import List, Callable, Union
import zstandard as zstd
import py7zr
import shutil
import psutil
from importlib.metadata import version, PackageNotFoundError
import multiprocessing as mp

from twb.logger import twb_logger


def get_curr_version():
    """
    Get the version of the package.
    """
    try:
        return version('twb-project')
    except PackageNotFoundError:
        return "Package not found"


def get_file_list(input_path: str) -> List[str]:
    """
    Get the list of files in the input directory.
    :param input_path: the input directory
    :raise FileNotFoundError: if the path does not exist
    :return: the list of files
    """
    # If the path does not exist, raise an error.
    if not os.path.exists(input_path):
        raise FileNotFoundError('The path does not exist.')

    # If the input path is a file, return the list with only the file path.
    if os.path.isfile(input_path):
        return [input_path]

    # If the input path is a directory, return the list of files in the directory.
    all_files = []
    for root, directories, files in os.walk(input_path):
        for file in files:
            if file.startswith('.'):
                continue
            file_path = os.path.join(root, file)
            all_files.append(file_path)

    # Remove duplicate files. Sort them for determinism.
    all_files = sorted(list(set(all_files)))
    return all_files


def get_decompress_output_path(input_path: str, output_dir: str):
    """
    Get the output path of the decompressed file.
    :param input_path: the input path
    :param output_dir: the output directory
    :return: the output path
    """
    # Split the path into its components
    _, file_name = os.path.split(input_path)

    # Might want to reconsider the output file structure.
    if not file_name.endswith('.7z'):
        return os.path.join(output_dir, file_name)

    decompressed_file_name = file_name.rstrip('.7z')

    # Add the new directory to the beginning of the path
    return os.path.join(output_dir, decompressed_file_name)


COMPRESSION_EXTENSION = '.zst'


def compress_zstd(input_path: str, output_path: str):
    """
    Compress the blocks into a Zstandard file.
    :param input_path: the input path
    :param output_path: the output path
    """
    # Compress the blocks.
    compressor = zstd.ZstdCompressor()
    with open(input_path, "rb") as ifh, open(output_path, "wb") as ofh:
        compressor.copy_stream(ifh, ofh)


def decompress_zstd(input_path: str, output_path: str):
    """
    Decompress the blocks from a Zstandard file.
    :param input_path: the input path
    :param output_path: the output path
    """
    # Decompress the blocks.
    decompressor = zstd.ZstdDecompressor()
    with open(input_path, "rb") as ifh, open(output_path, "wb") as ofh:
        decompressor.copy_stream(ifh, ofh)


def compute_total_available_space(output_dir: str) -> int:
    """
    Deprecated: Compute the total available space in the output directory.
    """
    total_available_space = shutil.disk_usage(output_dir).free

    # Display the total available space in GB.
    total_available_space_gb = total_available_space / 1024 / 1024 / 1024
    print('[Build] RDS space limitation (deprecated):', round(total_available_space_gb, 2), 'GB.')

    return total_available_space


def get_estimated_size(path: str) -> int:
    if path.endswith('.7z'):
        with py7zr.SevenZipFile(path, 'r') as z:
            space = z.archiveinfo().uncompressed
        return space * 2
    elif path.endswith('.zst'):
        with open(path, 'rb') as f:
            # Get the frame information for the compressed file
            space = zstd.frame_content_size(f.read(18))
            if space <= 0:
                return os.path.getsize(path) * 2
            else:
                return space * 2
    else:
        return os.path.getsize(path) * 2


def get_memory_consumption() -> int:
    process = psutil.Process(mp.current_process().pid)
    memory_usage_mb = process.memory_info().rss / 1024 / 1024
    return round(memory_usage_mb, 2)


def get_line_positions(path: str):
    """
    Get all line positions in the given file. So that it could be re-used to read the file for a specific line.
    """
    line_positions = []

    with open(path, 'r') as f:
        # Get position before reading the line so that it is the beginning of the line.
        position = f.tell()
        line = f.readline()
        while line:
            if len(line) > 0:
                line_positions.append(position)
            position = f.tell()
            line = f.readline()

    return line_positions


def read_line_in_file(path: str, position: int):
    """
    Read a specific line in the file without loading the entire file into memory.
    """
    with open(path, 'r') as f:
        f.seek(position)
        return f.readline()


def _rmtree_error_handler(func, path, exc_info):
    twb_logger.error(f"Error occurred while calling {func.__name__} on {path}")
    twb_logger.error(f"Error details: {exc_info}")

    # TODO: We might be able to attempt to resolve the issue based on exc_info and then retry the operation.


def cleanup_dir(path: str, onerror: Union[Callable, None] = _rmtree_error_handler):
    """
    Clean up the directory.
    :param path: the directory path
    :param onerror: the error handler
    """
    if os.path.exists(path):
        try:
            shutil.rmtree(path, onerror=onerror)
        except Exception as e:
            twb_logger.error(f"Error occurred while removing: {path}. Check next log for details.")
            twb_logger.error(e)


def prepare_output_dir(output_dir: str):
    if os.path.exists(output_dir):
        cleanup_dir(output_dir)
    os.makedirs(output_dir)


def parse_schema(obj):
    """
    Parse the schema of the given object. Used for glimpse.
    """
    if isinstance(obj, dict):
        if not obj:
            return "empty"
        return {key: parse_schema(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        if not obj:
            return "empty"
        return [parse_schema(obj[0]), len(obj)]
    else:
        return type(obj).__name__



from .builder import Builder
from .reader import Reader
from .downloader import Downloader
from .bip import BlockInteriorProcessor
from .decompressor import Decompressor
from .modifier import Modifier
